def parse_line(line)
  elms = []
  splitted = line.chomp.split("\t")
  splitted.each {|x|
    if x =~ /\d/
      elms << x.to_i
    else
      elms << false
    end
  }
  return elms
end
  
def read_file(filename)
  fruit_names = []
  matrix = []
  open(filename) {|file|
    line = file.gets
    fruit_names = line.chomp.split("\t")
    while line = file.gets
      matrix << parse_line(line)
    end
  }
  return fruit_names, matrix
end

def inner_prod(vec_a, vec_b)
  val = 0
  vec_a.zip(vec_b) {|a, b|
    val += a * b if a and b
  }
  return val
end

def vec_abs(vec)
  val = 0
  vec.each{|x|
    val += x * x if x
  }
  Math::sqrt(val)
end

def val_shift(vec)
  return vec.map{|x|
    if x then x - 3 else x end
  }
end

def cosine_similarity(vec_a, vec_b)
  vec_a = val_shift(vec_a)
  vec_b = val_shift(vec_b)

  nume = inner_prod(vec_a, vec_b)
  deno = vec_abs(vec_a) * vec_abs(vec_b)
  return nume / deno
end

def pearson_corr(vec_a, vec_b)
# まず平均を計算する
  mean_a = 0
  mean_b = 0
  n = 0.0
  vec_a.zip(vec_b).each{|a, b|
    if a and b
      n += 1
      mean_a += a
      mean_b += b
    end
  }

  if mean_a == 0 or mean_b == 0
    return 0
  end

  mean_a = mean_a / n
  mean_b = mean_b / n
  
  f = g = h = 0
  
  vec_a.zip(vec_b).each{|a, b|
    if a and b
      f += (a - mean_a) * (b - mean_b)
      g += (a - mean_a) ** 2
      h += (b - mean_b) ** 2
    end
  }
  p f, g, h
    return f / Math::sqrt(g * h)
end

def find_most_similar(vec_a, matrix)
  max_sim = 0
  sim_vec = false
  matrix.each{|vec_b|
    sim = cosine_similarity(vec_a, vec_b)
#   sim = pearson_corr(vec_a, vec_b)
    if sim > max_sim
      max_sim = sim
      sim_vec = vec_b
    end
  }
  return sim_vec
end

def recommend_for_user(vec, fruit_names, matrix)
  sim_vec = find_most_similar(vec, matrix)
  favor_fruits = []
  sim_vec.each_with_index {|elm, i|
    
    if elm and elm > 3 and vec[i] == false
      favor_fruits << [elm, fruit_names[i]]
    end
  }
  return favor_fruits.sort{|a, b| a[0] <=> b[0]}
end


def evaluate(filename, test_filename)
  fruit_names, matrix = read_file(filename)
  tp = tn = fp = fn = 0.0

  open(test_filename) {|file|
    line = file.gets # 1行めは名前なので読み飛ばす
    while line = file.gets
      vec = parse_line(line)
      r = rand(vec.length)
      hidden_item = [vec[r], fruit_names[r]]
      vec[r] = false

      result = recommend_for_user(vec, fruit_names, matrix)
      if (result.length > 0)

        # 1つしか隠してないので1つしか出ない
        if result[0][1] == hidden_item[1]
          if hidden_item[0] > 3
            tp += 1
          else
            fp += 1
          end
        end
      else
        if hidden_item[0] > 3
          fn += 1
        else
          tn += 1
        end
      end
    end
    return tp/(tp+fp), tp/(tp+fn)
  }
end

f = false
smith = [f, 5, 2, 3, 3, f, 3, 1]

filename = "data1.txt"
fruit_names, matrix = read_file(filename)
smith = [f, 5, 2, 3, 3, f, 3, 1]
favor_fruits = recommend_for_user(smith, fruit_names, matrix)
if favor_fruits.length == 0
  puts "No fruits found"
else
  favor_fruits.each{|val, name|
    puts "#{name} (#{val})"    
  }
end


precision, recall = evaluate("data1.txt", "data2.txt")
puts "precision = " + precision.to_s
puts "recall = " + recall.to_s

